﻿using System;
using System.IO;
using System.IO.Compression;
using System.Text;
using System.Linq;

namespace Ozeki
{
    public static class OzGZIP
    {
        static readonly byte[] GZipHeaderBytes = { 0x1f, 0x8b, 8, 0, 0, 0, 0, 0, 4, 0 };
        static readonly byte[] GZipLevel10HeaderBytes = { 0x1f, 0x8b, 8, 0, 0, 0, 0, 0, 2, 0 };

        public static bool IsGZIP(byte[] a)
        {
            var yes = a.Length > 10;

            if (!yes)
            {
                return false;
            }

            var header = a.Take(10).ToList();

            return header.SequenceEqual(GZipHeaderBytes) || header.SequenceEqual(GZipLevel10HeaderBytes);
        }

        public static string CompressToBase64(string text)
        {
            byte[] buffer = Encoding.UTF8.GetBytes(text);

            byte[] compressed = null;

            using (var ms = new MemoryStream())
            {
                using (var zip = new GZipStream(ms, CompressionMode.Compress, true))
                {
                    zip.Write(buffer, 0, buffer.Length);
                }

                ms.Position = 0;

                compressed = new byte[ms.Length];
                ms.Read(compressed, 0, compressed.Length);
            }

            byte[] gzBuffer = new byte[compressed.Length + 4];
            System.Buffer.BlockCopy(compressed, 0, gzBuffer, 4, compressed.Length);
            System.Buffer.BlockCopy(BitConverter.GetBytes(buffer.Length), 0, gzBuffer, 0, 4);

            return Convert.ToBase64String(gzBuffer);
        }

        public static string DecompressToBase64(string compressedText)
        {
            byte[] gzBuffer = Convert.FromBase64String(compressedText);
            using (MemoryStream ms = new MemoryStream())
            {
                int msgLength = BitConverter.ToInt32(gzBuffer, 0);
                ms.Write(gzBuffer, 4, gzBuffer.Length - 4);

                byte[] buffer = new byte[msgLength];

                ms.Position = 0;
                using (GZipStream zip = new GZipStream(ms, CompressionMode.Decompress))
                {
                    zip.Read(buffer, 0, buffer.Length);
                }

                return Encoding.UTF8.GetString(buffer);
            }
        }

        public static string CompressToString(string str, Encoding encoding = null)
        {
            if (encoding == null)
                encoding = Encoding.UTF8;

            var bytes = CompressToBytes(str);

            return encoding.GetString(bytes);
        }

        public static byte[] CompressToBytes(string str, Encoding encoding = null)
        {
            if (encoding == null)
                encoding = Encoding.UTF8;

            var bytes = encoding.GetBytes(str);

            return CompressToBytes(bytes);
        }

        public static bool CompressToStream(Stream memory, byte[] raw, bool leaveOpen)
        {
            try
            {
                using (var gzip = new GZipStream(memory, CompressionMode.Compress, leaveOpen))
                {
                    gzip.Write(raw, 0, raw.Length);
                }

                return true;
            }
            catch { return false; }
        }

        public static byte[] CompressToBytes(byte[] raw)
        {
            byte[] array = null;

            try
            {
                using (var memory = new MemoryStream())
                {
                    CompressToStream(memory, raw, false);

                    memory.Flush();
                    memory.Close();
                    array = memory.ToArray();
                }
            }
            catch (Exception e) { }

            return array;
        }

        public static string DecompressToString(byte[] gzip)
        {
            string sampleOut = null;

            var bytes = DecompressToBytes(gzip);

            sampleOut = Encoding.UTF8.GetString(bytes);

            return sampleOut;
        }

        public static byte[] DecompressToBytes(byte[] gzip)
        {
            byte[] array = null;

            using (var compressedStream = new MemoryStream(gzip))
            {
                using (var zipStream = new GZipStream(compressedStream, CompressionMode.Decompress, false))
                {
                    using (var resultStream = new MemoryStream())
                    {
                        zipStream.CopyTo(resultStream);
                        zipStream.Flush();
                        zipStream.Close();

                        array = resultStream.ToArray();
                    }
                }
            }

            return array;
        }

        public static MemoryStream DecompressToStream(Stream compressedStream)
        {
            var resultStream = new MemoryStream();

            using (var zipStream = new GZipStream(compressedStream, CompressionMode.Decompress, false))
            {
                zipStream.CopyTo(resultStream);
                zipStream.Flush();
                zipStream.Close();
            }

            resultStream.Position = 0;
            return resultStream;
        }


        public static bool DecompressToFile(Stream stream, string pathToDecompress, bool leaveOpen = false)
        {
            try
            {
                var dir = Path.GetDirectoryName(pathToDecompress);
                CreateDirIfNe(dir, out var err);
                using (var fs = File.Create(pathToDecompress))
                {
                    using (var gzip = new GZipStream(stream, CompressionMode.Decompress, leaveOpen))
                    {
                        gzip.CopyTo(fs);
                    }
                }
                return true;
            }
            catch (Exception)
            {
                return false;
            }
        }

        public static bool CreateDirIfNe(string dirname, out string errormessage)
        {
            errormessage = "";
            if (Directory.Exists(dirname)) return true;
            return CreateDir(dirname, out errormessage);
        }

        public static bool CreateDir(string dirname, out string errormessage)
        {
            errormessage = "";
            try
            {
                if (string.IsNullOrEmpty(dirname))
                {
                    errormessage = "Cannot create directory, directory name cannot NULL";
                    OzLogger.Log(OzLogger.DefaultLogSource, LogLevel.Error, errormessage);
                    return false;
                }

                if (Directory.Exists(dirname)) return true;
                Directory.CreateDirectory(dirname);
                return true;
            }
            catch (Exception exp)
            {
                errormessage = "Cannot create directory '" + dirname + "'." + exp.Message;
                OzLogger.Log(OzLogger.DefaultLogSource, LogLevel.Error, errormessage);
                return false;
            }
        }
    }
}