﻿//
// Megjegyzés:
// Az ebben a fájlban levő WebRequest-től, meg akarok szabadulni, ugyanis szerintem a WebClientEx mindent el tud intézni
// https://docs.microsoft.com/en-us/dotnet/api/system.net.http.httpclient?view=net-6.0
//
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Security;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Threading.Tasks;

namespace Ozeki
{
    static partial class OzHttpClient
    {
        //***********************************************************
        // Master HTTP download function
        //***********************************************************

        public static bool HttpDownload(OzHttpDownloadParams downloadParams)
        {
            if (downloadParams.PostData == null) downloadParams.PostData = string.Empty;
            if (LogRequests) OzLogger.Log(OzLogger.DefaultLogSource, LogLevel.Information, "HTTP client request, " + downloadParams.RequestMethod + ", " + downloadParams.Url);
            try
            {
                // setup request for execution
                var request = CreateRequest(downloadParams);

                // log request
                var logData = new StringBuilder();
                logRequest(downloadParams.RequestMethod, downloadParams.Url, extractResponseHeaders(request.Headers), downloadParams.PostData, logData);
                downloadParams.LogRequest = logData.ToString();

                // execute the request
                executeRequest(downloadParams, request, out var responseFirstLine, out var statusCode, out var responseHeaders, out var responseData, out var responseUrl);
                downloadParams.ResponseUrl = responseUrl;
                downloadParams.ResponseString = responseData;
                downloadParams.ResponseStatusCode = statusCode;

                // log request after response (the headers have changed)
                logData.Clear();
                logRequest(downloadParams.RequestMethod, downloadParams.Url, extractResponseHeaders(request.Headers), downloadParams.PostData, logData);
                downloadParams.LogRequest = logData.ToString();

                // log response
                logData.Clear();
                logResponse(responseFirstLine, responseHeaders.ToString(), responseData, logData);
                downloadParams.LogResponse = logData.ToString();

                return true;
            }
            catch (Exception e)
            {
                OzLogger.Log(OzLogger.DefaultLogSource, LogLevel.Warning, "HTTP download failed. " + downloadParams.RequestMethod.ToString() + " '" + downloadParams.Url + "' " + e.Message);
                downloadParams.ResponseString = string.Empty;
                downloadParams.LogResponse = string.Empty;
                downloadParams.Error = e.Message;
                return false;
            }
        }

        //***********************************************************
        // Create request
        //***********************************************************

        public static HttpWebRequest CreateRequest(OzHttpDownloadParams downloadParams)
        {
            setSecurityProtocol(downloadParams);

            HttpWebRequest request = null;
            switch (downloadParams.RequestMethod)
            {
                case HTTPRequestMethod.DELETE:
                    request = createPOSTRequest(downloadParams.Url);
                    request.Method = "DELETE";
                    break;
                case HTTPRequestMethod.GET:
                    request = createGETRequest(downloadParams.Url, downloadParams.PostData);
                    break;
                case HTTPRequestMethod.POST:
                    request = createPOSTRequest(downloadParams.Url);
                    break;
                case HTTPRequestMethod.PUT:
                    request = createPOSTRequest(downloadParams.Url);
                    request.Method = "PUT";
                    break;
            }

            SetupRequestHeaders(downloadParams, request);

            var proxyConfig = downloadParams.ProxyConfig;
            if (proxyConfig != null)
            {
                attachProxyConfig(request, proxyConfig.Proxy);
                setServerCertificate(request, proxyConfig.HttpsClient);
                setClientCertificate(request, proxyConfig.HttpsClient);
            }


            if (downloadParams.DisableCertificateValidation)
                disableCertificateValidation();

            // Ennek a legvégén kell lennie, mert lezárja a headerekt.
            if (needPostData(downloadParams))
            {
                attachPostData(request, downloadParams.PostData);
            }

            if (downloadParams.ProxyConfig != null && downloadParams.ProxyConfig.Proxy != null && 
                !string.IsNullOrEmpty(downloadParams.ProxyConfig.Proxy.ProxyHost) &&
                downloadParams.ProxyConfig.Proxy.ProxyPort != 0)
            {
                request.Proxy = new WebProxy("http://" + proxyConfig.Proxy.ProxyHost + ":" + proxyConfig.Proxy.ProxyPort);
            }

            if (downloadParams.TimeoutInSec > 2) 
            {
                request.Timeout = downloadParams.TimeoutInSec*1000;
            }

            return request;
        }

        static bool needPostData(OzHttpDownloadParams downParams)
        {
            if (downParams.PostData == null) return false;
            switch (downParams.RequestMethod)
            {
                case HTTPRequestMethod.GET:
                    return false;
                case HTTPRequestMethod.POST:
                case HTTPRequestMethod.PUT:
                case HTTPRequestMethod.DELETE:
                default:
                    return true;
            }
        }

        static void SetupRequestHeaders(OzHttpDownloadParams downloadParams, HttpWebRequest request)
        {
            setupOzekiHeaders(downloadParams);
            if (downloadParams.RequestHeaders.ContainsKey("User-Agent"))
                request.UserAgent = downloadParams.RequestHeaders["User-Agent"];
            if (downloadParams.RequestHeaders.ContainsKey("Accept"))
                request.Accept = downloadParams.RequestHeaders["Accept"];
            if (downloadParams.RequestHeaders.ContainsKey("Content-Type"))
                request.ContentType = downloadParams.RequestHeaders["Content-Type"];
            if (downloadParams.RequestHeaders.ContainsKey("Referer"))
                request.Referer = downloadParams.RequestHeaders["Referer"];
            if (downloadParams.RequestHeaders.ContainsKey("Host"))
                request.Host = downloadParams.RequestHeaders["Host"];
            if (downloadParams.RequestHeaders.ContainsKey("Connection"))
            {
                request.Connection = downloadParams.RequestHeaders["Connection"];
                request.KeepAlive = false;
            }
            setupCustomHeaders(request.Headers, downloadParams, true);
            setupHeaderAuthorization(request, downloadParams);
            SetupHeaders(request, downloadParams.RequestHeaders);
        }

        //***********************************************************
        // Create request / GET
        //***********************************************************

        static HttpWebRequest createGETRequest(string targetURLBase, string requestData)
        {
            string dataToSend = targetURLBase;
            if (requestData.Length > 0)
                dataToSend += "?" + requestData;
            var request = WebRequest.CreateHttp(dataToSend);
            request.Method = "GET";
            request.Timeout = 12000;
            return request;
        }

        //***********************************************************
        // Create request / POST
        //***********************************************************

        static HttpWebRequest createPOSTRequest(string targetURLBase)
        {
            var request = WebRequest.CreateHttp(targetURLBase);
            request.ContentType = "application/x-www-form-urlencoded; charset=UTF-8";
            request.Method = "POST";
            request.Timeout = 12000;
            return request;
        }

        //***********************************************************
        // Extract response headers
        //***********************************************************

        static Dictionary<string, string> extractResponseHeaders(WebHeaderCollection headers)
        {
            var headerDict = new Dictionary<string, string>();
            if (headers == null) return headerDict;
            if (headers.Count == 0) return headerDict;
            for (int i = 0; i < headers.Count; ++i)
            {
                string header = headers.GetKey(i);
                foreach (string value in headers.GetValues(i))
                {
                    if (!headerDict.ContainsKey(header))
                        headerDict.Add(header, value);
                    else
                        headerDict[header] += "; " + value;
                }
            }
            return headerDict;
        }

        //***************************************************************************
        // Private support functions
        //***************************************************************************

        static void executeRequest(OzHttpDownloadParams downloadParams, HttpWebRequest request, out string responseFirstLine, out HttpStatusCode statusCode, out WebHeaderCollection responseHeaders, out string responseData, out string responseUrl)
        {
            try
            {
                var response = OzWebRequestExecutor.Instance.ExecuteRequest(request) as HttpWebResponse;
                readResponse(downloadParams, response, out responseFirstLine, out statusCode, out responseHeaders, out responseData, out responseUrl);
            }
            catch (WebException webEx) // Ha sikertelen állapotot jelöl a státuszkód, akkor kivételt dob a WebRequest.GetResponse() függvény.
            {
                var response = webEx.Response as HttpWebResponse;
                readResponse(downloadParams, response, out responseFirstLine, out statusCode, out responseHeaders, out responseData, out responseUrl);
            }
        }

        static void readResponse(OzHttpDownloadParams downloadParams, HttpWebResponse response, out string firstLine, out HttpStatusCode statusCode, out WebHeaderCollection headers, out string responseData, out string responseUrl)
        {
            if (response == null)
            {
                firstLine = "HTTP/1.1 404 Not Found";
                statusCode = HttpStatusCode.NotFound;
                headers = new WebHeaderCollection();
                responseData = string.Empty;
                responseUrl = downloadParams.Url;
                return;
            }

            downloadParams.ResponseHeaders = extractResponseHeaders(response.Headers);

            firstLine = "HTTP/" + response.ProtocolVersion + " " + ((int)response.StatusCode).ToString() + " " + response.StatusDescription;
            statusCode = response.StatusCode;
            var receiveStream = response.GetResponseStream();
            var enc = Encoding.UTF8;
            if (downloadParams.ResponseEncodingIsInHttpHeader)
            {
                if (response.CharacterSet != null)
                {
                    try
                    {
                        enc = Encoding.GetEncoding(response.CharacterSet);
                    }
                    catch
                    {
                        enc = Encoding.UTF8;
                    }
                }
            }

            Stream rStream = receiveStream;

            if (downloadParams.ResponseHeaders.ContainsKey("Content-Encoding") && downloadParams.ResponseHeaders["Content-Encoding"] == "gzip")
            {
                rStream = OzGZIP.DecompressToStream(receiveStream);
            }

            using (var readStream = new StreamReader(rStream, enc))
            {
                responseData = readStream.ReadToEnd();
                readStream.Close();
            }
            headers = response.Headers;
            response.Close();
            responseUrl = response.ResponseUri.ToString();
        }
    }
}
