﻿using System;
using System.Linq;

namespace Ozeki
{
    public static class OzLogger
    {
        public static OzLogSource DefaultLogSource = OzLogSource.Create("Program");

        static readonly char[] _separator = new char[] { '\n' };
        static readonly char[] _trimChars = new char[] { '\r' };

        //*****************************************
        // public Log functions
        //*****************************************

        public static void Log(OzLogSource logSource, LogLevel level, string logMessage)
        {
            Log(logSource, level, string.Empty, logMessage);
        }

        public static void Log(OzLogSource[] logSources, LogLevel level, string logMessage)
        {
            foreach (var logSource in logSources.Distinct())
                Log(logSource, level, string.Empty, logMessage);
        }

        public static void Log(OzLogSource[] logSources, LogLevel level, string prefix, string logMessage, bool ignorePrefixOnFirstLine = true, bool includeEmptyLines = false)
        {
            foreach (var logSource in logSources.Distinct())
                Log(logSource, level, prefix, logMessage, ignorePrefixOnFirstLine, includeEmptyLines);
        }

        public static void Log(OzLogSource logSource, LogLevel level, string prefix, string logMessage, bool ignorePrefixOnFirstLine = true, bool includeEmptyLines = false)
        {
            if (string.IsNullOrEmpty(logMessage)) return;
            if (logSource == null) return;

            logMessage = logMessage.Replace('\0', ' ');

            if (!logMessage.Contains("\n"))
            {
                passToSystemLogger(logSource, level, prefix, logMessage, ignorePrefixOnFirstLine);
                return;
            }

            //Break log lines when new line is found
            var l = logMessage.Split(_separator);
            string logstring = string.Empty;
            for (int i = 0; i < l.Length; i++)
            {
                logstring = l[i].TrimEnd(_trimChars);
                if (!includeEmptyLines && logstring.Length <= 0) continue;
                passToSystemLogger(logSource, level, prefix, logstring, i == 0 && ignorePrefixOnFirstLine, includeEmptyLines);
            }
        }

        public static void Log(OzLogSource logSource, OzLogMessage logMessage)
        {
            logMessage.LogSource = logSource;
            CallLogWrite(logSource, logMessage);
        }

        static void passToSystemLogger(OzLogSource sender, LogLevel level, string prefix, string logMessage, bool ignorePrefixOnFirstLine, bool includeEmptyLines = false)
        {
            OzLogMessage logmsg;
            if (ignorePrefixOnFirstLine)
                logmsg = new OzLogMessage(sender, level, string.Empty, logMessage, includeEmptyLines);
            else
                logmsg = new OzLogMessage(sender, level, prefix, logMessage, includeEmptyLines);

            CallLogWrite(sender, logmsg);
        }

        //*****************************************
        // Logwrite event
        //*****************************************

        public static event EventHandler<OzLoggerEventArgs> OnLogWrite;

        static void CallLog(OzLogSource sender, OzLogMessage message)
        {
            OnLogWrite(sender, new OzLoggerEventArgs(message));
        }

        static void CallLogWrite(OzLogSource sender, OzLogMessage message)
        {
            if (sender.previousMsg != null)
            {
                if (message.EqualsMessage(sender.previousMsg))
                {
                    sender._count++;
                    sender.previousMsg = message;
                    return;
                }
                else
                {
                    if (sender._count > 1)
                    {
                        sender.previousMsg.LogMessage += " (" + sender._count + " times)";
                        CallLog(sender, sender.previousMsg);
                        sender._count = 0;
                        sender.previousMsg = null;
                    }
                    else if (sender._count == 1)
                    {
                        sender._count = 0;
                        CallLog(sender.previousMsg.LogSource, sender.previousMsg);
                    }
                }
            }
            sender.previousMsg = message;
            CallLog(sender, message);
        }

        //*****************************************
        // Reset
        //*****************************************

        public static void Reset(OzLogSource logSource)
        {
            logSource.previousMsg = null;
            logSource._count = 0;
            CallOnReset(logSource);
        }

        public static event EventHandler OnReset;
        static void CallOnReset(OzLogSource sender)
        {
            OnReset(sender, EventArgs.Empty);
        }
    }
}